
#include "mc_832.h"
#include "motor.h"
#include "../msim.h"
#include "../msimutil.h"


#include <assert.h>
#include <stdlib.h>    // rand, abs
#include <limits.h>
#include <stdio.h>     // sprintf
#include <string.h>    // strcat
#include <math.h>      // ceil
#include <iomanip.h>


MC_832* theMC_832;

#define LOG(cmds) \
    if(m_controller->GetLogStream() && m_controller->GetLogLevel()>0) { \
      (*m_controller->GetLogStream()) << cmds; \
    }

#define COMMAND(cmds) \
    LOG("<cmd c='C832'>" << cmds << "</cmd>" << endl);

#define OUTPUT(cmds)  \
    LOG("<return src='sim'>" << cmds << "</return>" << endl);

#define LOGERROR_(cmds) \
    if (m_motor->m_controller->GetLogStream()) { \
      *(m_motor->m_controller->GetLogStream()) \
        << "<error c='C832'>" \
        << cmds \
        << "</error>" \
        << endl; \
    }
///////////////////////////////////////////////////////////////////////////////
// class MC_832

MC_832::MC_832( )
{
  m_log_stream = 0;
  m_log_level = 1;

  m_addr_reg_addr = 0x210;
  m_data_reg_addr = 0x211;
  m_target_register = cmd_reg;

  m_motor[ 0 ] = new MC_832::Motor( this, 0 );
  m_motor[ 1 ] = new MC_832::Motor( this, 1 );

  m_current_motor = 0;

  m_starttime = timeGetTime();
}

MC_832::~MC_832( )
{
  delete m_motor[ 0 ];
  delete m_motor[ 1 ];
}

//inline
void MC_832::UpdateStatus( )
{
  updateSimulationStatusWindow( );
}

void MC_832::Put( unsigned address, byte value )
{
  if( address == m_data_reg_addr )
  {
    // value written to data register

    switch ( m_target_register )
    {
      case cmd_reg:
        m_motor[ m_current_motor ] -> PutCmd( value );
        break;
      case data_reg:
        m_motor[ m_current_motor ] -> PutData( value );
        break;
    };

  }
  else if ( address == m_addr_reg_addr )
  {
    // value written to address register
    if ( ( value & 0x07 ) == 0x07 )
    {
      m_target_register = int_reg;
    }
    else
    {
      m_current_motor   = ( value & 2 ) ? 1 : 0;
      m_target_register = ( value & 1 ) ? data_reg : cmd_reg;
    }
  }
}

byte MC_832::Get( unsigned address )
{
  if( address == m_data_reg_addr )
  {

    switch ( m_target_register )
    {
      case cmd_reg:
        return m_motor[ m_current_motor ] -> GetStat( );
      case data_reg:
        return m_motor[ m_current_motor ] -> GetData( );
      case int_reg:
        // "read and reset interrupt source"
        if( GetLogStream( ) && GetLogLevel( ) >0 )
        {
          ( *GetLogStream( ) )
              << "<cmd c='C832'>"
                 "'read and reset interrupt source'"
                 "</cmd>" << endl;
        }
        byte reg = 0;
        reg |= m_motor[ 0 ] -> HasLimitHit( ) ? 1 : 0;
        reg |= m_motor[ 1 ] -> HasLimitHit( ) ? 2 : 0;
        m_motor[ 0 ] -> SetLimitHit( false );
        m_motor[ 1 ] -> SetLimitHit( false );
        return reg;
    }

  }
  return 0;
}

const char* MC_832::GetStatusText( )
{
  m_motor[ 0 ] -> CalcPosition( );
  m_motor[ 1 ] -> CalcPosition( );
  UpdateStatusText( );
  return m_status_text;
}

void MC_832::UpdateStatusText( )
{
  char tmp[ 100 ];
  int i;

  m_status_text[ 0 ]  = 0;
  for( i=0; i<2; i++ )
  {
    sprintf( tmp,
             "%cM%i: %li, %li, %li \r\n   [%i%i%i%i%i%i%i%i %i%i%i%i%i%i%i%i] %c\r\n",
             m_motor[i] -> m_last_direction ? '>' : '<',
             i+1,
             m_motor[i] -> m_position,
             m_motor[i] -> m_physical_position,
             m_motor[i] -> m_target - m_motor[i] ->m_position,
             ( m_motor[i] -> m_signals_register & 32768L )  == 32768L,
             ( m_motor[i] -> m_signals_register & 16384 )  == 16384,
             ( m_motor[i] -> m_signals_register & 8192 )  == 8192,
             ( m_motor[i] -> m_signals_register & 4096 )  == 4096,
             ( m_motor[i] -> m_signals_register & 2048 )  == 2048,
             ( m_motor[i] -> m_signals_register & 1024 )  == 1024,
             ( m_motor[i] -> m_signals_register & 512 )  == 512,
             ( m_motor[i] -> m_signals_register & 256 )  == 256,

             ( m_motor[i] -> m_signals_register & 128 )  == 128,
             ( m_motor[i] -> m_signals_register & 64 )  == 64,
             ( m_motor[i] -> m_signals_register & 32 )  == 32,
             ( m_motor[i] -> m_signals_register & 16 )  == 16,
             ( m_motor[i] -> m_signals_register & 8 )  == 8,
             ( m_motor[i] -> m_signals_register & 4 )  == 4,
             ( m_motor[i] -> m_signals_register & 2 )  == 2,
             ( m_motor[i] -> m_signals_register & 1 )  == 1,

             (m_motor[i] -> HasLimitHit() ? '!' : ' ')
           );
    strcat( m_status_text, tmp );
  }
}


///////////////////////////////////////////////////////////////////////////////
// class MC_832::Motor

MC_832::Motor::Motor( MC_832* controller, int board_id )
{
  m_controller = controller;
  m_board_id = board_id;

  m_status = 0;
  m_signals_register = 0;
  SetMotorOff( true );
  SetVelocityMode( true );
  SetOnTarget( true );

  m_current_cmd = 0;
  m_current_trj = 0;
  m_current_fil = 0;

  m_acceleration = 0;
  m_velocity = 0;
  m_velocity_max = 0;

  m_target = 0;
  m_position = 0;
  m_position_time = 0;
  m_index_position = 0;

  m_physical_position_range = LONG_MAX;
  m_physical_position = LONG_MAX / 2;
  m_physical_index_position = 0;

  m_hysteresis = 0;
  m_last_direction = true;

  m_has_limit_hit = false;

  m_kp = m_ki = m_kd = m_kl = 0;
}

void MC_832::Motor::PutData( byte b )
{
  if ( m_current_cmd && m_current_cmd->AcceptsPut( ) )
  {
    m_current_cmd->Put( b );
  }
  else
  {
    // ??
  }

}

void MC_832::Motor::PutCmd( byte b )
{
  if ( m_current_cmd )
  {
    if ( m_current_cmd->CMD( ) != CMD_LTRJ ) delete m_current_cmd;
    m_current_cmd = 0;
  }
  // dispatch the command
  switch ( b )
  {
    case CMD_RESET:
      RESET( );
      break;
    case CMD_RDSIGS:
      RDSIGS( );
      break;
    case CMD_DFH:
      DFH( );
      break;
    case CMD_LTRJ:
      LTRJ( );
      break;
    case CMD_STT:
      STT( );
      break;
    case CMD_RDRP:
      RDRP( );
      break;
    case CMD_RDDP:
      RDDP( );
      break;
    case CMD_LFIL:
      LFIL( );
      break;
    case CMD_UDF:
      UDF( );
      break;
    case CMD_RDIP:
      RDIP( );
      break;
    case CMD_SIP:
      SIP( );
      break;
    case CMD_MSKI:
      MSKI( );
      break;
    case CMD_RSTI:
      RSTI( );
      break;
  }

  m_controller -> UpdateStatus( );
}

byte MC_832::Motor::GetData( )
{
  if ( m_current_cmd && m_current_cmd->AcceptsGet( ) )
  {
    return m_current_cmd->Get( );
  }
  else
  {
    // ???
    return 0;
  }
}

byte MC_832::Motor::GetStat( )
{
  COMMAND( "'get status' " << m_board_id );
  CalcPosition( );
  m_controller -> UpdateStatus( );
  return m_status;
}

void MC_832::Motor::RESET( )
{
  COMMAND( "reset LM628 " << m_board_id );
  // define home
  CalcPosition( );
  m_position = 0;
  // reset trj. params
  m_target = 0;
  m_velocity = 0;
  m_velocity_max = 0;
  m_acceleration = 0;
  SetOnTarget( true );
  // clear the interrupt masks (bit 0-5 are set to 1, bit 6 remains unchanged)
  m_interrupt_mask |= 0x3F;

  if ( m_current_fil ) delete m_current_fil;
  m_current_fil = 0;
  if ( m_current_trj ) delete m_current_trj;
  m_current_trj = 0;

  m_kp = m_ki = m_kd = m_kl = 0;
}

void MC_832::Motor::RDSIGS( )
{
  COMMAND( "read signals register " << m_board_id );
  CalcPosition( );
  m_current_cmd = new MC_832::CmdRDSIGS( this, m_signals_register );
  OUTPUT( binary( m_signals_register ) );
}

void MC_832::Motor::DFH( )
{
  COMMAND( "define home " << m_board_id );

  CalcPosition( );
  m_position = 0;

  // ??
  // "if executed during motion it will not affect the stopping position ..."
}


void MC_832::Motor::LTRJ( )
{
  COMMAND( "load trajectory " << m_board_id );

  if ( !m_current_trj )
  {
    // create new trajectory command
    m_current_trj = new MC_832::CmdLTRJ( this );
  }
  else
  {
    MC_832::CmdLTRJ* old = m_current_trj;
    // create new trajectory command, which holds infos in current
    // trajectory buffer to
    m_current_trj = new MC_832::CmdLTRJ( this, *m_current_trj );
    delete old;
  }
  m_current_cmd = m_current_trj;

}

void MC_832::Motor::STT( )
{
  short ctrl;

  LOG( "<cmd c='C832'>" << "start motion " << m_board_id << " " );

  if ( m_current_trj )
  {
    // excute
    CalcPosition( );
    ctrl = m_current_trj -> m_ctrl.short_val;

    // velocity mode at bit 11
    SetVelocityMode( ( ctrl & 2048 ) == 2048 );

    if ( ctrl & 1024 )
    {
      // bit 10
      LOG( "stop smoothly" );
      StopSmoothly();
    }

    else if ( ctrl & 512 )
    {
      // bit 9
      LOG( "stop abruptly" );
      StopAbruptly( );
    }
    else if ( ctrl & 256 )
    {
      // bit 8
      LOG( "motor off" );
      SetMotorOff( true );
    }
    else
    {
      SetMotorOff( false );

      ctrl = m_current_trj -> m_dataflags;

      // acceleration
      if ( ctrl & MC_832::CmdLTRJ::accel_loaded )
      {
        if ( ctrl & MC_832::CmdLTRJ::accel_relative )
        {
          // relative values (only positive values are allowed!?)
          m_acceleration += m_current_trj -> m_accel.long_val;
          LOG( "ar: " << m_current_trj -> m_accel.long_val << " " );
        }
        else
        {
          m_acceleration = m_current_trj -> m_accel.long_val;
          LOG( "aa: " << m_current_trj -> m_accel.long_val << " " );
        }
        SetAccelerationLoaded( false );
      }
      // velocity
      if ( ctrl & MC_832::CmdLTRJ::velo_loaded )
      {
        if ( ctrl & MC_832::CmdLTRJ::velo_relative )
        {
          // relative values (only positive values are allowed!?)
          m_velocity_max += m_current_trj -> m_velo.long_val;
          LOG( "vr: " << m_current_trj -> m_velo.long_val << " " );
        }
        else
        {
          m_velocity_max = m_current_trj -> m_velo.long_val;
          LOG( "va: " << m_current_trj -> m_velo.long_val << " " );
        }
      }
      // position
      if ( ctrl & MC_832::CmdLTRJ::pos_loaded )
      {
        if ( ctrl & MC_832::CmdLTRJ::pos_relative )
        {
          m_target = m_position + m_current_trj -> m_pos.long_val;
          LOG( "pr: " << m_current_trj -> m_pos.long_val << " " );
        }
        else
        {
          m_target = m_current_trj -> m_pos.long_val;
          LOG( "pa: " << m_current_trj -> m_pos.long_val << " " );
        }

        // direction changed ?
        if ( m_last_direction == right && m_target < m_position )
        {
          m_last_direction = left;
          m_physical_position += m_hysteresis;
          // if we change the direction we have to stop
          m_velocity = 0;
        }
        else if ( m_last_direction == left && m_target > m_position )
        {
          m_last_direction = right;
          m_physical_position -= m_hysteresis;
          // if we change the direction we have to stop
          m_velocity = 0;
        }

        SetOnTarget( m_target == m_position );
      }
    }
    // forget
    delete m_current_trj;
    m_current_trj = 0;
  }
  else
  {
    LOG( "<warn>STT without LTRJ</warn>" );
  }

  LOG( "</cmd>" << endl );
}

void MC_832::Motor::RDRP( )
{
  CalcPosition( );
  m_current_cmd = new MC_832::CmdRDRP( this, m_position );

  COMMAND( "read real position " << m_board_id );
  OUTPUT( m_position );
}

void MC_832::Motor::RDDP( )
{
  CalcPosition( );
  m_current_cmd = new MC_832::CmdRDDP( this, m_target );

  COMMAND( "read desired position " << m_board_id );
  OUTPUT( m_target );
}

void MC_832::Motor::LFIL( )
{
  COMMAND( "load filter parameters " << m_board_id );

  m_current_cmd = new MC_832::CmdLFIL( this );
}

void MC_832::Motor::UDF( )
{
  COMMAND( "update filter " << m_board_id );
}

void MC_832::Motor::RDIP( )
{
  COMMAND( "read index position " << m_board_id );
  OUTPUT( m_index_position );

  m_current_cmd = new MC_832::CmdRDIP( this, m_index_position );
}

void MC_832::Motor::SIP( )
{
  COMMAND( "set index position " << m_board_id );

  SetIndexFound( false );
  SetWaitingForIndex( true );
}

void MC_832::Motor::MSKI( )
{
  COMMAND( "mask interrupts " << m_board_id );

  m_current_cmd = new MC_832::CmdMSKI( this );
}

void MC_832::Motor::RSTI( )
{
  COMMAND( "reset interrupts " << m_board_id );

  m_current_cmd = new MC_832::CmdRSTI( this );
}





long MC_832::Motor::CalcPosition( )
{
  return CalcPosition_Ideal( );
}

#define _sq(x)        ((x) * (x))
#define _error        (m_target - m_position + \
                      ((m_target>=m_position) ? (long)-ceil(s_1) : (long)ceil(s_1)))
#define _error_       (m_target - m_position)
#define _decel_dist   (_sq(v_1) / (2 * m_acceleration))
#define _decel_dist_  (_sq(m_velocity) / (2 * m_acceleration))

long MC_832::Motor::CalcPosition_Ideal( )
{
  long t = m_controller -> GetCurrentTime( );
  double delta_t = ( t - m_position_time )  / 1000.0;
  double s_1, s_2, v_1, v_2;

  // check that we are really moving
  if ( IsMotorOff( ) )  { goto end; }
  if ( m_target == m_position )  { SetOnTarget( true ); goto end; }
  if ( ! ( delta_t > 0.0 )  )  goto end;

  SetOnTarget( false );

  if (  labs( _error_ )  > _decel_dist_ )
  {

    if ( m_velocity < m_velocity_max )
    {
      // in phase 1
      s_1 = ( m_acceleration / 2 )  * _sq( delta_t )  +
            m_velocity * delta_t;
      v_1 = m_acceleration * delta_t + m_velocity;

      if ( labs(_error) > _decel_dist )
      {

        if ( v_1 < m_velocity_max )
        {
          // stay in phase 1
          s_2 = s_1;
          v_2 = v_1;
        }
        else
        {
          // change to phase 2
          s_2 = m_velocity_max * delta_t;
          s_2 = fabs( s_1 + s_2 )  / 2;
          v_2 = m_velocity_max;
        }
        goto check_target_reached;
      }
      else
      {
        goto change_to_phase_3;
      }
    }
    else
    {
      // in Phase 2
      s_1 = m_velocity_max * delta_t;
      v_1 = m_velocity_max;

      if (  labs( _error )  > _decel_dist )
      {
        // stay in phase 2
        s_2 = s_1;
        v_2 = m_velocity_max;
        goto check_target_reached;
      }
      else
      {
        goto change_to_phase_3;
      }
    }
  }
  else
  {
    // in phase 3
    s_2 = ( -m_acceleration / 2 )  * _sq( delta_t )  +
             m_velocity * delta_t;
    v_2 = ( -m_acceleration * delta_t )  + m_velocity;
    goto check_target_reached;
  }


change_to_phase_3:
  {
    s_2 = ( -m_acceleration / 2 )  * _sq( delta_t )  +
            m_velocity * delta_t;
    s_2 = fabs( s_1 + s_2 ) / 2;
    v_2 = ( -m_acceleration * delta_t )  + m_velocity;
    v_2 = fabs( v_1 + v_2 ) / 2;
  }

check_target_reached:
  {
    if ( m_position <= m_target )
    {
      // move direction -->
      m_position += (long)ceil( s_2 );
      m_physical_position += (long)ceil( s_2 );

      if ( IsWaitingForIndex( ) )
      {
        if ( between( m_physical_index_position,
                      m_physical_position - (long)ceil( s_2 ),
                      m_physical_position) )
        {
          m_index_position = m_position
                - labs( m_physical_position - m_physical_index_position );
          SetWaitingForIndex( false );
          SetIndexFound( true );
        }
      }

      if ( m_physical_position > m_physical_position_range )
      {
        // hardware limit switch reached
        // the motor doesn't stop but signals the hit in the int_reg
        SetLimitHit( true );
      }
      else
      {
        SetLimitHit( false );
      }

      m_velocity = (m_velocity < v_2) ? (long)ceil( v_2 ) : (long)floor( v_2 );
      if ( ! ( m_position < m_target && v_2 > 0.0 )  )
      {
        // target reached or over shot
        m_physical_position -= ( m_position - m_target );
        m_position = m_target;
        // stop
        m_velocity = 0;
      }
    }
    else
    {
      // move direction <--
      m_position -= (long)ceil( s_2 );
      m_physical_position -= (long)ceil( s_2 );

      if ( IsWaitingForIndex( ) )
      {
        if ( between( m_physical_index_position,
                      m_physical_position,
                      m_physical_position + (long)ceil( s_2 ) ) )
        {
          m_index_position = m_position
                + labs( m_physical_position - m_physical_index_position );
          SetWaitingForIndex( false );
          SetIndexFound( true );
        }
      }

      if ( m_physical_position < 0 )
      {
        // hardware limit switch reached
        // the motor doesn't stop but signals the hit in the int_reg
        SetLimitHit( true );
      }
      else
      {
        SetLimitHit( false );
      }

      m_velocity = (m_velocity < v_2) ? (long)ceil( v_2 ) : (long)floor( v_2 );
      if ( ! ( m_position > m_target && v_2 > 0.0 )  )
      {
        // target reached or over shot
        m_physical_position += ( m_target - m_position );
        m_position = m_target;
        // stop
        m_velocity = 0;
      }
    }
  }

  if( m_controller->GetLogStream() )
  {
    ( *(m_controller->GetLogStream()) ) << setbase( 10 )
        << "<info c='c832'>time " << t << " position " << m_position
        << "/" << m_physical_position
        << " velocity " << m_velocity << "</info>" << endl;
  }


end:
  {
    if ( delta_t > 0.0 )  m_position_time = t;

    return m_position;
  }
}

void MC_832::Motor::StopSmoothly( )
{
  CalcPosition( );

  // way needed to stop with programmed deceleration (==acceleration)
  long decel_dist = _sq( m_velocity )  / ( 2 * m_acceleration );

  // set target to the closest point to reach if deceleration starts now
  if ( m_last_direction == right )
  {
    m_target = m_position + decel_dist;
  }
  else
  {
    m_target = m_position - decel_dist;
  }
}

void MC_832::Motor::StopAbruptly( )
{
  CalcPosition( );
  m_target = m_position;
  m_velocity = 0;
  SetOnTarget( true );
}




///////////////////////////////////////////////////////////////////////////////
// class MC_832::Command

MC_832::Command::Command( MC_832::Motor* m )  : m_motor( m )
{}


MC_832::Command::~Command( )
{
  m_motor = 0;
}
///////////////////////////////////////////////////////////////////////////////
// class MC_832::CmdLTRJ

MC_832::CmdLTRJ::CmdLTRJ( MC_832::Motor *m )  : MC_832::Command( m )
{
  m_accel.long_val = 0;
  m_velo.long_val = 0;
  m_pos.long_val = 0;

  m_ctrl.short_val = 0;
  m_state = ctrl_byte1;
  m_dataflags = 0;
}

MC_832::CmdLTRJ::CmdLTRJ( MC_832::Motor *m, MC_832::CmdLTRJ& ltrj )
    : MC_832::Command( m )
{
  m_accel.long_val = ltrj.m_accel.long_val;
  m_velo.long_val = ltrj.m_velo.long_val;
  m_pos.long_val = ltrj.m_pos.long_val;
  m_dataflags = ltrj.m_dataflags;

  m_ctrl.short_val = 0;
  m_state = ctrl_byte1;
}

/*
MC_832::CmdLTRJ::CmdLTRJ( MC_832::CmdLTRJ& ltrj )
    : MC_832::Command( m_motor )
{
  m_accel.long_val = ltrj.m_accel.long_val;
  m_velo.long_val = ltrj.m_velo.long_val;
  m_pos.long_val = ltrj.m_pos.long_val;
  m_dataflags = ltrj.m_dataflags;

  m_ctrl.short_val = 0;
  m_state = ctrl_byte1;
}*/

bool MC_832::CmdLTRJ::AcceptsPut( )
{
  return ( m_state != end );
}

bool MC_832::CmdLTRJ::AcceptsGet( )
{
  return false;
}

byte MC_832::CmdLTRJ::Get( )
{
  assert( false );
  return 0;
}

void MC_832::CmdLTRJ::Put( byte b )
{
  switch ( m_state )
  {
    case ctrl_byte1:
      m_ctrl.bytes[ 1 ]  = b;
      m_state = ctrl_byte0;
      break;
    case ctrl_byte0:
      m_ctrl.bytes[ 0 ]  = b;
      switch ( m_ctrl.short_val & 0x2A ) // 101010
      {
        case 0x20:
        case 0x22:
        case 0x28:
        case 0x2A:                       // bit 5 set
          m_state = accel_byte3;
          break;
        case 0x08:
        case 0x0A:                       // bit 3 set
          m_state = velo_byte3;
          break;
        case 0x02:                       // bit 1 set
          m_state = pos_byte3;
          break;
        default:
          m_state = end;
      };
      break;

    case accel_byte3:
      m_accel.bytes[ 3 ]  = b;
      m_state = accel_byte2;
      break;
    case accel_byte2:
      m_accel.bytes[ 2 ]  = b;
      m_state = accel_byte1;
      break;
    case accel_byte1:
      m_accel.bytes[ 1 ]  = b;
      m_state = accel_byte0;
      break;
    case accel_byte0:
      m_accel.bytes[ 0 ]  = b;
      m_dataflags |= accel_loaded;
      m_dataflags |= ( m_ctrl.short_val & 16 ) ? accel_relative : 0;

      if (! between (m_accel.long_val, 0L, 0x3FFFFFFFL))
      {
        LOGERROR_("LTRJ: acceleration value outside valid range [0..3FFFFFFF]");
      }
      m_motor -> SetAccelerationLoaded( true );

      switch ( m_ctrl.short_val & 0x0A )
      {                                // 1010
        case 0x08:
        case 0x0A:
          m_state = velo_byte3;
          break;
        case 0x02:
          m_state = pos_byte3;
          break;
        default:
          m_state = end;
      };
      break;

    case velo_byte3:
      m_velo.bytes[ 3 ]  = b;
      m_state = velo_byte2;
      break;
    case velo_byte2:
      m_velo.bytes[ 2 ]  = b;
      m_state = velo_byte1;
      break;
    case velo_byte1:
      m_velo.bytes[ 1 ]  = b;
      m_state = velo_byte0;
      break;
    case velo_byte0:
      m_velo.bytes[ 0 ]  = b;
      m_dataflags |= velo_loaded;
      m_dataflags |= ( m_ctrl.short_val & 4 )  ? velo_relative : 0;

      if (! between (m_velo.long_val, 0L, 0x3FFFFFFFL))
      {
        LOGERROR_("LTRJ: velocity value outside valid range [0..3FFFFFFF]");
      }

      switch ( m_ctrl.short_val & 0x02 )
      { // 10
        case 0x02:
          m_state = pos_byte3;
          break;
        default:
          m_state = end;
      };
      break;
    case pos_byte3:
      m_pos.bytes[ 3 ]  = b;
      m_state = pos_byte2;
      break;
    case pos_byte2:
      m_pos.bytes[ 2 ]  = b;
      m_state = pos_byte1;
      break;
    case pos_byte1:
      m_pos.bytes[ 1 ]  = b;
      m_state = pos_byte0;
      break;
    case pos_byte0:
      m_pos.bytes[ 0 ]  = b;
      m_dataflags |= pos_loaded;
      m_dataflags |= ( m_ctrl.short_val & 1 ) ? pos_relative : 0;

      if (! between (m_pos.long_val, LONG_MIN, 0x3FFFFFFFL))
      {
        LOGERROR_("LTRJ: position value outside valid range [-C0000000..3FFFFFFF], actually [LONG_MIN....3FFFFFFF]");
      }

      m_state = end;
      break;
    default:
      assert( false );
  }
}

///////////////////////////////////////////////////////////////////////////////
// class MC_832::CmdMSKI

MC_832::CmdMSKI::CmdMSKI( MC_832::Motor* m ) : MC_832::Command( m )
{
  m_state = data_byte1;
  m_data.short_val = 0;
}

bool MC_832::CmdMSKI::AcceptsPut( )
{
  return ( m_state != end );
}

bool MC_832::CmdMSKI::AcceptsGet( )
{
  return false;
}

byte MC_832::CmdMSKI::Get( )
{
  assert( false );
  return 0;
}

void MC_832::CmdMSKI::Put( byte b )
{
  switch ( m_state )
  {
    case data_byte1:
      m_data.bytes[ 1 ]  = b;
      m_state = data_byte0;
      break;
    case data_byte0:
      m_data.bytes[ 0 ]  = b;
      m_state = end;
      break;
  }
  if ( m_state == end )
  {
    // execute
    m_motor -> SetInterruptMask( m_data.short_val );
  }
}


///////////////////////////////////////////////////////////////////////////////
// class MC_832::CmdRSTI

bool MC_832::CmdRSTI::AcceptsPut( )
{
  return ( m_state != end );
}

bool MC_832::CmdRSTI::AcceptsGet( )
{
  return false;
}

byte MC_832::CmdRSTI::Get( )
{
  assert( false );
  return 0;
}

void MC_832::CmdRSTI::Put( byte b )
{
  switch ( m_state )
  {
    case data_byte1:
      m_data.bytes[ 1 ]  = b;
      m_state = data_byte0;
      break;
    case data_byte0:
      m_data.bytes[ 0 ]  = b;
      m_state = end;
      break;
  }

  if ( m_state == end )
  {
    // execute
    m_motor -> ResetInterruptMask( m_data.short_val );

    // some status bits are cleard by RSTI
    m_motor -> SetBreakpointReached( false );
    m_motor -> SetExcessivePositionError( false );
    m_motor -> SetWraparoundObserved( false );
    m_motor -> SetIndexFound( false );
    m_motor -> SetTrjComplete( false );
    m_motor -> SetCmdError( false );
  }
}


///////////////////////////////////////////////////////////////////////////////
// class MC_832::CmdLFIL

bool MC_832::CmdLFIL::AcceptsPut( )
{
  return ( m_state != end );
}

bool MC_832::CmdLFIL::AcceptsGet( )
{
  return false;
}

byte MC_832::CmdLFIL::Get( )
{
  assert( false );
  return 0;
}

void MC_832::CmdLFIL::Put( byte b )
{
  switch ( m_state )
  {
    case ctrl_byte1:
      m_ctrl.bytes[ 1 ] = b;
      m_state = ctrl_byte0;
      break;
    case ctrl_byte0:
      m_ctrl.bytes[ 0 ] = b;
      if ( m_ctrl.bytes[ 0 ] & 8 ) { m_state = kp_byte1; break; }
      if ( m_ctrl.bytes[ 0 ] & 4 ) { m_state = ki_byte1; break; }
      if ( m_ctrl.bytes[ 0 ] & 2 ) { m_state = kd_byte1; break; }
      if ( m_ctrl.bytes[ 0 ] & 1 ) { m_state = il_byte1; break; }
      // else
      m_state = end;
      break;

    case kp_byte1:
      m_kp.bytes[ 1 ] = b;
      m_state = kp_byte0;
      break;
    case kp_byte0:
      m_kp.bytes[ 0 ] = b;

      if (! between(m_kp.short_val, (short)0, (short)0x7FFF) )
      {
        LOGERROR_("LFIL: kp value out of valid range [0..7fff]");
      }

      if ( m_ctrl.bytes[ 0 ] & 4 ) { m_state = ki_byte1; break; }
      if ( m_ctrl.bytes[ 0 ] & 2 ) { m_state = kd_byte1; break; }
      if ( m_ctrl.bytes[ 0 ] & 1 ) { m_state = il_byte1; break; }
      // else
      m_state = end;
      break;

    case ki_byte1:
      m_ki.bytes[ 1 ] = b;
      m_state = ki_byte0;
      break;
    case ki_byte0:
      m_ki.bytes[ 0 ] = b;

      if (! between(m_ki.short_val, (short)0, (short)0x7FFF) )
      {
        LOGERROR_("LFIL: ki value out of valid range [0..7fff]");
      }

      if ( m_ctrl.bytes[ 0 ] & 2 ) { m_state = kd_byte1; break; }
      if ( m_ctrl.bytes[ 0 ] & 1 ) { m_state = il_byte1; break; }
      // else
      m_state = end;
      break;

    case kd_byte1:
      m_kd.bytes[ 1 ] = b;
      m_state = kd_byte0;
      break;
    case kd_byte0:
      m_kd.bytes[ 0 ] = b;

      if (! between(m_kd.short_val, (short)0, (short)0x7FFF) )
      {
        LOGERROR_("LFIL: kd value out of valid range [0..7fff]");
      }

      if ( m_ctrl.bytes[ 0 ] & 1 ) { m_state = il_byte1; break; }
      // else
      m_state = end;
      break;

    case il_byte1:
      m_il.bytes[ 1 ] = b;
      m_state = il_byte0;
      break;
    case il_byte0:
      m_il.bytes[ 0 ] = b;

      if (! between(m_il.short_val, (short)0, (short)0x7FFF) )
      {
        LOGERROR_("LFIL: il value out of valid range [0..7fff]");
      }

      m_state = end;
      break;

  }
}

///////////////////////////////////////////////////////////////////////////////
// class MC_832::Cmd32BitReport

bool MC_832::Cmd32BitReport::AcceptsPut( )
{
  return false;
}

bool MC_832::Cmd32BitReport::AcceptsGet( )
{
  return ( m_state != end );
}

byte MC_832::Cmd32BitReport::Get( )
{
  switch ( m_state )
  {
    case rpt_byte3:
      m_state = rpt_byte2;
      return m_value.bytes[ 3 ];
    case rpt_byte2:
      m_state = rpt_byte1;
      return m_value.bytes[ 2 ];
    case rpt_byte1:
      m_state = rpt_byte0;
      return m_value.bytes[ 1 ];
    case rpt_byte0:
      m_state = end;
      return m_value.bytes[ 0 ];
    default:
      assert( false );
      return 0;
  }
}

void MC_832::Cmd32BitReport::Put( byte )
{
  assert( false );
}


///////////////////////////////////////////////////////////////////////////////
// class MC_832::Cmd16BitReport

bool MC_832::Cmd16BitReport::AcceptsPut( )
{
  return false;
}

bool MC_832::Cmd16BitReport::AcceptsGet( )
{
  return ( m_state != end );
}

byte MC_832::Cmd16BitReport::Get( )
{
  switch ( m_state )
  {
    case rpt_byte1:
      m_state = rpt_byte0;
      return m_value.bytes[ 1 ];
    case rpt_byte0:
      m_state = end;
      return m_value.bytes[ 0 ];
    default:
      assert( false );
      return 0;
  }
}

void MC_832::Cmd16BitReport::Put( byte )
{
  assert( false );
}

///////////////////////////////////////////////////////////////////////////////

void MC_832::SetHysteresis( int board_id, long value )
{
  assert( 0 <= board_id && board_id <= 1 );
  m_motor[ board_id ] -> m_hysteresis = value;
}

long MC_832::GetHysteresis( int board_id )
{
  assert( 0 <= board_id && board_id <= 1 );
  return m_motor[ board_id ] -> m_hysteresis;
}

void MC_832::SetLastDirection( int board_id, bool direction )
{
  assert( 0 <= board_id && board_id <= 1 );
  m_motor[ board_id ] -> m_last_direction = direction;
}

bool MC_832::GetLastDirection( int board_id )
{
  assert( 0 <= board_id && board_id <= 1 );
  return (m_motor[ board_id ] -> m_last_direction) ? true : false ;
}

void MC_832::SetPhysicalPositionRange( int board_id, long val )
{
  assert( 0 <= board_id && board_id <= 1 );
  m_motor[ board_id ] -> m_physical_position_range = val;
}

long MC_832::GetPhysicalPositionRange( int board_id )
{
  assert( 0 <= board_id && board_id <= 1 );
  return m_motor[ board_id ] -> m_physical_position_range;
}

void MC_832::SetPhysicalPosition( int board_id, long val )
{
  assert( 0 <= board_id && board_id <= 1 );
  m_motor[ board_id ] -> m_physical_position = val;
}

long MC_832::GetPhysicalPosition( int board_id )

{
  assert( 0 <= board_id && board_id <= 1 );
  return m_motor[ board_id ] -> m_physical_position;
}

void MC_832::SetPhysicalIndexPosition( int board_id, long val )
{
  assert( 0 <= board_id && board_id <= 1 );
  m_motor[ board_id ] -> m_physical_index_position = val;
}

long MC_832::GetPhysicalIndexPosition( int board_id )
{
  assert( 0 <= board_id && board_id <= 1 );
  return m_motor[ board_id ] -> m_physical_index_position;
}


