
#ifndef MC_832_MOTOR_H
#define MC_832_MOTOR_H

const byte CMD_RESET  = 0x00;
const byte CMD_PORT8  = 0x05;
const byte CMD_PORT12 = 0x06;
const byte CMD_DFH    = 0x02;
const byte CMD_SIP    = 0x03;
const byte CMD_LPEI   = 0x1B;
const byte CMD_LPES   = 0x1A;
const byte CMD_SBPA   = 0x20;
const byte CMD_SBPR   = 0x21;
const byte CMD_MSKI   = 0x1C;
const byte CMD_RSTI   = 0x1D;
const byte CMD_LFIL   = 0x1E;
const byte CMD_UDF    = 0x04;
const byte CMD_LTRJ   = 0x1F;
const byte CMD_STT    = 0x01;
const byte CMD_RDSIGS = 0x0C;
const byte CMD_RDIP   = 0x09;
const byte CMD_RDDP   = 0x08;
const byte CMD_RDRP   = 0x0A;
const byte CMD_RDDV   = 0x07;
const byte CMD_RDRV   = 0x0B;
const byte CMD_RDSUM  = 0x0D;


class MC_832::Motor
{
    friend class MC_832;
    friend class MC_832::CmdLTRJ;
    friend class MC_832::CmdMSKI;
    friend class MC_832::CmdRSTI;
    friend class MC_832::CmdLFIL;
    friend class MC_832::CmdRDIP;
    friend class MC_832::CmdRDRP;
    friend class MC_832::CmdRDSIGS;

  public:
    MC_832::Motor( MC_832* controller, int board_id );

    void RESET();
    void RDSIGS();

    void DFH();
    void LTRJ();
    void STT();
    void RDRP();
    void RDDP();

    void LFIL();
    void UDF();

    void RDIP();
    void SIP();

    void MSKI();
    void RSTI();

    void PutData( byte b );
    void PutCmd( byte b );
    byte GetData();
    byte GetStat();

  protected:
    // set status bits
    void SetMotorOff( bool on );
    bool IsMotorOff();
    void SetBreakpointReached( bool on );
    void SetExcessivePositionError( bool on );
    void SetWraparoundObserved( bool on );
    void SetIndexFound( bool on );
    void SetTrjComplete( bool on );
    void SetCmdError( bool on );
    void SetBusy( bool on );
    void SetOnTarget( bool on );
    bool IsOnTarget();
    void SetVelocityMode( bool on );
    bool IsInVelocityMode();
    void SetWaitingForIndex( bool on );
    bool IsWaitingForIndex();
    void SetAccelerationLoaded( bool on );
    void SetUDFExecuted( bool on );

    bool HasLimitHit();
    void SetLimitHit( bool );

    long CalcPosition();
    long CalcPosition_Ideal();
    void StopSmoothly();
    void StopAbruptly();

    void SetInterruptMask( short val );
    void ResetInterruptMask( short val );

  private:
    MC_832* m_controller;
    int m_board_id;

    MC_832::Command* m_current_cmd;
    // trajectory input buffer
    MC_832::CmdLTRJ* m_current_trj;
    // filter input buffer
    MC_832::CmdLFIL* m_current_fil;

    byte m_status;
    short m_signals_register;
    // 1 at bit x means that interrupt x is raised (theoretically)
    short m_interrupt_mask;

    enum { velocity_mode, position_mode };

    long m_acceleration;
    // programmed velocity
    long m_velocity_max;
    // actual velocity at m_position_time
    long m_velocity;


    // position to be hold/reached
    long m_target;
    // actual position at m_position_time
    long m_position;
    long m_position_time;
    long m_index_position;

    long m_physical_position_range;
    long m_physical_position;
    long m_physical_index_position;
    long m_hysteresis;
    enum { left = false, right = true };
    int m_last_direction;

    bool m_has_limit_hit;

    // coefficents for the proportional integral derivative (PID)
    // filter which controlls the power of the motor control output
    // (in simulation currently not used)
    long m_kp, m_ki, m_kd, m_kl;

};


class MC_832::Command
{
  public:
    MC_832::Command( MC_832::Motor* m );
    virtual ~Command();

    // indicates whether the command is waiting for a written data byte
    // report commands return always false
    virtual bool AcceptsPut() = 0;
    // indicates whether the command provides a byte to be read
    virtual bool AcceptsGet() = 0;
    // writes the the byte to a control command
    // should not be called if AcceptsPut returns false
    virtual void Put( byte ) = 0;
    // returns the next byte of a report command
    // should not be called if AcceptsGet returns false
    virtual byte Get() = 0;
    // returns the command code of the command for identifying purposes
    virtual int  CMD() = 0;

  protected:
    MC_832::Motor* m_motor;
};


class MC_832::CmdLTRJ : public MC_832::Command
{
    friend class MC_832::Motor;

    enum state {
      ctrl_byte1,  // byte 1 of 16 bit control word expected
      ctrl_byte0,
      accel_byte3, // 3rd byte of 32 bit acceleration parameter expected
      accel_byte2,
      accel_byte1,
      accel_byte0,
      velo_byte3,
      velo_byte2,  // 2nd byte of 32 bit velocity parameter expected
      velo_byte1,
      velo_byte0,
      pos_byte3,
      pos_byte2,
      pos_byte1,
      pos_byte0,
      end
  };


  public:
    MC_832::CmdLTRJ( MC_832::Motor *m );
    MC_832::CmdLTRJ( MC_832::Motor *m, MC_832::CmdLTRJ& ltrj );
    //MC_832::CmdLTRJ( MC_832::CmdLTRJ& ltrj );

    virtual bool AcceptsPut();
    virtual bool AcceptsGet();
    virtual void Put( byte );
    virtual byte Get();
    virtual int  CMD()  { return CMD_LTRJ; }

    enum {
      accel_loaded = 32,
      accel_relative = 16,
      velo_loaded = 8,
      velo_relative = 4,
      pos_loaded   = 2,
      pos_relative = 1
    };

  protected:
    state m_state;
    // stores the contol word
    short_bytes m_ctrl;
    // indicates which trajectory parameters are transmitted
    // since last STT command
    int m_dataflags;
    // store the trajectory parameters
    long_bytes m_accel;
    long_bytes m_velo;
    long_bytes m_pos;
};



class MC_832::CmdMSKI : public MC_832::Command
{
    enum state {
      data_byte1,   // byte 1 of the 16 bit data word expected
      data_byte0,
      end
  };
  public:
    MC_832::CmdMSKI( MC_832::Motor* m );

    virtual bool AcceptsPut();
    virtual bool AcceptsGet();
    virtual void Put( byte );
    virtual byte Get();
    virtual int  CMD()  { return CMD_MSKI; }

  private:
    state m_state;
    short_bytes m_data;
};

class MC_832::CmdRSTI : public MC_832::Command
{
    enum state {
      data_byte1,   // byte 1 of the 16 bit data word expected
      data_byte0,
      end
  };
  public:
    MC_832::CmdRSTI( MC_832::Motor* m )  : MC_832::Command( m )
    {
      m_state = data_byte1;
      m_data.short_val = 0;
    }

    virtual bool AcceptsPut();
    virtual bool AcceptsGet();
    virtual void Put( byte );
    virtual byte Get();
    virtual int  CMD()  { return CMD_RSTI; }

  private:
    state m_state;
    short_bytes m_data;
};


class MC_832::CmdLFIL : public MC_832::Command
{
    enum state {
      ctrl_byte1,   // byte 1 of the 16 bit data word expected
      ctrl_byte0,
      kp_byte3,
      kp_byte2,
      kp_byte1,
      kp_byte0,
      ki_byte3,
      ki_byte2,
      ki_byte1,
      ki_byte0,
      kd_byte3,
      kd_byte2,
      kd_byte1,
      kd_byte0,
      il_byte3,
      il_byte2,
      il_byte1,
      il_byte0,
      end
  };
  public:
    MC_832::CmdLFIL( MC_832::Motor* m )  : MC_832::Command( m )
    {
      m_state = ctrl_byte1;

      m_ctrl.short_val = 0;
      m_kp.short_val = 0;
      m_ki.short_val = 0;
      m_kd.short_val = 0;
      m_il.short_val = 0;
    }

    virtual bool AcceptsPut();
    virtual bool AcceptsGet();
    virtual void Put( byte );
    virtual byte Get();
    virtual int  CMD()  { return CMD_LFIL; }

  private:
    state m_state;

    short_bytes m_ctrl;
    short_bytes m_kp;
    short_bytes m_ki;
    short_bytes m_kd;
    short_bytes m_il;
};


class MC_832::Cmd32BitReport : public MC_832::Command
{
    enum state {
      rpt_byte3,
      rpt_byte2,
      rpt_byte1,   // byte 1 of the 32 bit data word provided
      rpt_byte0,
      end
  };
  public:
    MC_832::Cmd32BitReport( MC_832::Motor* m, long value ) : MC_832::Command( m )
    {
      m_state = rpt_byte3;
      m_value.long_val = value;
    }

    virtual bool AcceptsPut();
    virtual bool AcceptsGet();
    virtual void Put( byte );
    virtual byte Get();

  private:
    state m_state;
    long_bytes m_value;
};


class MC_832::CmdRDRP : public MC_832::Cmd32BitReport
{
  public:
    MC_832::CmdRDRP( MC_832::Motor* m, long pos )
       : MC_832::Cmd32BitReport( m, pos )
    {}

    virtual int  CMD()  { return CMD_RDRP; }
};


class MC_832::CmdRDIP : public MC_832::Cmd32BitReport
{
  public:
    MC_832::CmdRDIP( MC_832::Motor* m, long pos )
        : MC_832::Cmd32BitReport( m, pos )
    {}

    virtual int  CMD()  { return CMD_RDIP; }
};

class MC_832::CmdRDDP : public MC_832::Cmd32BitReport
{
  public:
    MC_832::CmdRDDP( MC_832::Motor* m, long pos )
        : MC_832::Cmd32BitReport( m, pos )
    {}

    virtual int  CMD()  { return CMD_RDDP; }
};

class MC_832::Cmd16BitReport : public MC_832::Command
{
    enum state {
      rpt_byte1,   // byte 1 of the 16 bit data word provided
      rpt_byte0,
      end
  };
  public:
    MC_832::Cmd16BitReport( MC_832::Motor* m, short value )  : MC_832::Command( m )
    {
      m_state = rpt_byte1;
      m_value.short_val = value;
    }

    virtual bool AcceptsPut();
    virtual bool AcceptsGet();
    virtual void Put( byte );
    virtual byte Get();

  private:
    state m_state;
    short_bytes m_value;
};


class MC_832::CmdRDSIGS : public MC_832::Cmd16BitReport
{
  public:
    MC_832::CmdRDSIGS( MC_832::Motor* m, short sigs )
    : MC_832::Cmd16BitReport( m, sigs )
    {}

    virtual int  CMD()  { return CMD_RDSIGS; }
};




inline
void MC_832::Motor::SetMotorOff( bool on )
{
  // bit 7 in status register and in signals reg.
  if ( on )
  {
    m_status |= 128;
    m_signals_register |= 128;
    SetTrjComplete( true );
  }
  else
  {
    m_status &= ~128;
    m_signals_register &= ~128;
    SetTrjComplete( IsOnTarget() );
  }
}

inline
bool MC_832::Motor::IsMotorOff()
{
  return ( ( m_status & 128 )  == 128 )  ? true : false;
}


inline
void MC_832::Motor::SetBreakpointReached( bool on )
{
  // bit 6 in status register and in signals reg.
  if ( on )
  {
    m_status |= 64;
    m_signals_register |= 64;
  }
  else
  {
    m_status &= ~64;
    m_signals_register &= ~64;
  }
}

inline
void MC_832::Motor::SetExcessivePositionError( bool on )
{
  // bit 5 in status register and in signals reg.
  if ( on )
  {
    m_status |= 32;
    m_signals_register |= 32;
  }
  else
  {
    m_status &= ~32;
    m_signals_register &= ~32;
  }
}

inline
void MC_832::Motor::SetWraparoundObserved( bool on )
{
  // bit 4 in status register and in signals reg.
  if ( on )
  {
    m_status |= 16;
    m_signals_register |= 16;
  }
  else
  {
    m_status &= ~16;
    m_signals_register &= ~16;
  }
}


inline
void MC_832::Motor::SetIndexFound( bool on )
{
  // bit 3 in status register and in signals reg.
  if ( on )
  {
    m_status |= 8;
    m_signals_register |= 8;
  }
  else
  {
    m_status &= ~8;
    m_signals_register &= ~8;
  }
}

inline
void MC_832::Motor::SetWaitingForIndex( bool on )
{
  // bit 0 in signals reg.
  if ( on )
  {
    m_signals_register |= 1;
  }
  else
  {
    m_signals_register &= ~1;
  }
}

inline
bool MC_832::Motor::IsWaitingForIndex()
{
  // bit 0 in signals reg.
  return ( ( m_signals_register & 1 )  == 1 )  ? true : false;
}

inline
void MC_832::Motor::SetTrjComplete( bool on )
{
  // bit 2 in status register and in signals reg.
  if ( on )
  {
    m_status |= 4;
    m_signals_register |= 4;
  }
  else
  {
    m_status &= ~4;
    m_signals_register &= ~4;
  }
}

inline
void MC_832::Motor::SetCmdError( bool on )
{
  // bit 1 in status register and in signals reg.
  if ( on )
  {
    m_status |= 2;
    m_signals_register |= 2;
  }
  else
  {
    m_status &= ~2;
    m_signals_register &= ~2;
  }
}

inline
void MC_832::Motor::SetBusy( bool on )
{
  // bit 0
  if ( on ) 	m_status |= 1;
  else	m_status &= ~1;
}

inline
void MC_832::Motor::SetOnTarget( bool on )
{
  // bit 10 in signals reg.
  if ( on )
  {
    m_signals_register |= 1024;
    SetTrjComplete( true );
  }
  else
  {
    m_signals_register &= ~1024;
    SetTrjComplete( IsMotorOff() );
  }
}

inline
bool MC_832::Motor::IsOnTarget()
{
  return ( m_status & 1024 )  == 1024;
}

inline
void MC_832::Motor::SetVelocityMode( bool on )
{
  // bit 11 in signals reg.
  if( on )
  {
    m_signals_register |= 2048;
  }
  else
  {
    m_signals_register &= ~2048;
  }
}

inline
bool MC_832::Motor::IsInVelocityMode()
{
  return ( m_status & 2048 )  == 2048;
}

inline
void MC_832::Motor::SetAccelerationLoaded( bool on )
{
  // bit 14 in signals reg.
  if( on )
  {
    m_signals_register |= 16384;
  }
  else
  {
    m_signals_register &= ~16384;
  }
}

inline
void MC_832::Motor::SetUDFExecuted( bool on )
{
  // bit 13 in signals reg.
  if( on )
  {
    m_signals_register |= 8192;
  }
  else
  {
    m_signals_register &= ~8192;
  }
}

inline
void MC_832::Motor::SetInterruptMask( short mask )
{
  m_interrupt_mask = mask;
}

inline
void MC_832::Motor::ResetInterruptMask( short mask )
{
  // ???
  m_interrupt_mask = mask;
}

inline
bool MC_832::Motor::HasLimitHit()
{
  CalcPosition();
  return m_has_limit_hit;
}

inline
void MC_832::Motor::SetLimitHit( bool val )
{
  m_has_limit_hit = val;
}

#endif MC_832_MOTOR_H

